"""
Detectors package.
"""

from detectors.eval import get_ood_results

from . import config, data, eval, methods, models, utils
from .aggregations import create_aggregation, list_aggregations, register_aggregation
from .data import create_dataset, get_dataset_cls, list_datasets, register_dataset
from .methods import Detector, create_detector, create_hyperparameters, list_detectors, register_detector
from .methods.utils import create_reduction
from .models import create_model, create_transform, list_models, timm_register_model
from .pipelines import create_pipeline, list_pipelines, register_pipeline
from .aggregations import create_aggregation, list_aggregations, register_aggregation
