from detectors.data.cifar_wrapper import CIFAR10_LABELS, CIFAR100_LABELS
from detectors.data.cifarc import CORRUPTIONS

CIFAR10_DEFAULT_MEAN = (0.4914, 0.4822, 0.4465)
CIFAR10_DEFAULT_STD = (0.2023, 0.1994, 0.2010)
CIFAR100_DEFAULT_MEAN = (0.5071, 0.4867, 0.4408)
CIFAR100_DEFAULT_STD = (0.2675, 0.2565, 0.2761)
SVHN_DEFAULT_MEAN = (0.4377, 0.4438, 0.4728)
SVHN_DEFAULT_STD = (0.1980, 0.2010, 0.1970)
STL10_DEFAULT_MEAN = (0.43, 0.42, 0.39)
STL10_DEFAULT_STD = (0.27, 0.26, 0.27)
MNIST_DEFAULT_MEAN = (0.173,)
MNIST_DEFAULT_STD = (0.332,)
IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)
